@@ -3,12 +3,14 @@ import collections
 import threading
 from http import HTTPStatus
 from io import BytesIO
-from typing import Any, Coroutine, Deque, Iterable, Optional
+from typing import Any, Coroutine, Deque, Iterable, Optional, TypeVar
 from typing import cast as typing_cast
 
 from .asgi_typing import HTTPScope, ASGIApp, ReceiveEvent, SendEvent
 from .wsgi_typing import Environ, StartResponse, IterableChunks
 
+T = TypeVar("T")
+
 
 class defaultdict(dict):
     def __init__(self, default_factory, *args, **kwargs) -> None:
@@ -21,7 +23,7 @@ class defaultdict(dict):
 
 StatusStringMapping = defaultdict(
     lambda status: f"{status} Unknown Status Code",
-    {int(status): f"{status} {status.phrase}" for status in HTTPStatus},
+    {status.value: f"{status.value} {status.phrase}" for status in HTTPStatus},
 )
 
 
@@ -152,11 +154,13 @@ class ASGIResponder:
         self.wait_time = wait_time
 
         self.sync_event = SyncEvent()
-        self.async_event = AsyncEvent(loop)
-        self.async_lock: asyncio.Lock
+        self.sync_event_set_lock: asyncio.Lock
+
+        self.receive_event = AsyncEvent(loop)
+        self.send_event = AsyncEvent(loop)
 
         def _init_async_lock():
-            self.async_lock = asyncio.Lock()
+            self.sync_event_set_lock = asyncio.Lock()
 
         loop.call_soon_threadsafe(_init_async_lock)
 
@@ -164,14 +168,14 @@ class ASGIResponder:
         self.wsgi_should_stop: bool = False
 
     async def asgi_receive(self) -> ReceiveEvent:
-        async with self.async_lock:
-            self.sync_event.set({"type": "receive"})
-            return await self.async_event.wait()
+        await self.sync_event_set_lock.acquire()
+        self.sync_event.set({"type": "receive"})
+        return await self.receive_event.wait()
 
     async def asgi_send(self, message: SendEvent) -> None:
-        async with self.async_lock:
-            self.sync_event.set(message)
-            await self.async_event.wait()
+        await self.sync_event_set_lock.acquire()
+        self.sync_event.set(message)
+        await self.send_event.wait()
 
     def asgi_done_callback(self, future: asyncio.Future) -> None:
         try:
@@ -180,20 +184,23 @@ class ASGIResponder:
             pass
         else:
             if exception is not None:
-                self.sync_event.set(
-                    {
-                        "type": "a2wsgi.error",
-                        "exception": (
-                            type(exception),
-                            exception,
-                            exception.__traceback__,
-                        ),
-                    }
+                task = asyncio.create_task(self.sync_event_set_lock.acquire())
+                task.add_done_callback(
+                    lambda _: self.sync_event.set(
+                        {
+                            "type": "a2wsgi.error",
+                            "exception": (
+                                type(exception),
+                                exception,
+                                exception.__traceback__,
+                            ),
+                        }
+                    )
                 )
         finally:
             self.asgi_done.set()
 
-    def start_asgi_app(self, environ: Environ) -> asyncio.Task:
+    async def start_asgi_app(self, environ: Environ) -> asyncio.Task:
         run_asgi: asyncio.Task = self.loop.create_task(
             typing_cast(
                 Coroutine[None, None, None],
@@ -203,19 +210,25 @@ class ASGIResponder:
         run_asgi.add_done_callback(self.asgi_done_callback)
         return run_asgi
 
+    def execute_in_loop(self, coro: Coroutine[None, None, T]) -> T:
+        return asyncio.run_coroutine_threadsafe(coro, self.loop).result()
+
     def __call__(
         self, environ: Environ, start_response: StartResponse
     ) -> IterableChunks:
         read_count: int = 0
         body = environ["wsgi.input"] or BytesIO()
         content_length = int(environ.get("CONTENT_LENGTH", None) or 0)
+        receive_eof = False
+        body_sent = False
 
-        asgi_task = self.start_asgi_app(environ)
+        asgi_task = self.execute_in_loop(self.start_asgi_app(environ))
         # activate loop
         self.loop.call_soon_threadsafe(lambda: None)
 
         while True:
             message = self.sync_event.wait()
+            self.loop.call_soon_threadsafe(self.sync_event_set_lock.release)
             message_type = message["type"]
 
             if message_type == "http.response.start":
@@ -230,13 +243,21 @@ class ASGIResponder:
                     ],
                     None,
                 )
+                self.send_event.set(None)
             elif message_type == "http.response.body":
                 yield message.get("body", b"")
+                body_sent = True
                 self.wsgi_should_stop = not message.get("more_body", False)
+                self.send_event.set(None)
             elif message_type == "http.response.disconnect":
                 self.wsgi_should_stop = True
+                self.send_event.set(None)
             # ASGI application error
             elif message_type == "a2wsgi.error":
+                if body_sent:
+                    raise message["exception"][1].with_traceback(
+                        message["exception"][2]
+                    )
                 start_response(
                     "500 Internal Server Error",
                     [
@@ -248,32 +269,33 @@ class ASGIResponder:
                 yield b"Server got itself in trouble"
                 self.wsgi_should_stop = True
             elif message_type == "receive":
-                pass
-            else:
-                raise RuntimeError(f"Unknown message type: {message_type}")
-
-            if message_type == "receive":
                 read_size = min(65536, content_length - read_count)
-                data: bytes = body.read(read_size)
-                if data == b"":
-                    self.async_event.set({"type": "http.disconnect"})
+                if read_size == 0:  # No more body, so don't read anymore
+                    if not receive_eof:
+                        self.receive_event.set(
+                            {"type": "http.request", "body": b"", "more_body": False}
+                        )
+                        receive_eof = True
+                    else:
+                        pass  # let `await receive()` wait
                 else:
+                    data: bytes = body.read(read_size)
                     read_count += len(data)
                     more_body = read_count < content_length
-                    self.async_event.set(
+                    self.receive_event.set(
                         {"type": "http.request", "body": data, "more_body": more_body}
                     )
             else:
-                self.async_event.set(None)
+                raise RuntimeError(f"Unknown message type: {message_type}")
 
             if self.wsgi_should_stop:
-                self.async_event.set_nowait()
+                self.receive_event.set({"type": "http.disconnect"})
                 break
 
-            if asgi_task.done():
+            if self.asgi_done.is_set():
                 break
 
         # HTTP response ends, wait for run_asgi's background tasks
         self.asgi_done.wait(self.wait_time)
-        asgi_task.cancel()
+        self.loop.call_soon_threadsafe(asgi_task.cancel)
         yield b""
